import React, {
  ChangeEvent,
  JSXElementConstructor,
  useCallback,
  useEffect,
  useMemo,
  useState,
} from "react";
import {
  GridColDef,
  GridRowsProp,
  DataGridPro,
  GridToolbarContainer,
  GridToolbarFilterButton,
  GridFilterModel,
  GridFilterOperator,
  GridFilterItem,
  GridFilterInputValueProps,
  GridState,
  GridLinkOperator,
  GridSelectionModel,
} from "@mui/x-data-grid-pro";
import {
  GetAction,
  QueryBuilder,
  withNetworkCompletionDispatch,
} from "@nerdjs/nerd-network";
import { ActionCreatorWithPreparedPayload, Dispatch } from "@reduxjs/toolkit";
import StringOperatorInput from "./customOperatorInput/stringTypeInput";
import DateOperatorInput from "./customOperatorInput/dateTypeInput";
import BoolOperatorInput from "./customOperatorInput/boolTypeInput";
import CustomPagination from "../NerdPagination/NerdPagination";
import { Button } from "@mui/material";
import { SaveAlt } from "@mui/icons-material";

const OPERATORS = {
  eq: {
    label: "Equal",
    value: "eq",
  },
  neq: {
    label: "Not Equal",
    value: "neq",
  },
  gt: {
    label: "Greater Than",
    value: "gt",
  },
  lt: {
    label: "Less Than",
    value: "lt",
  },
  gte: {
    label: "Greater or Equal Than",
    value: "gte",
  },
  lte: {
    label: "Less or Equal Than",
    value: "lte",
  },
  like: {
    label: "Like",
    value: "like",
  },
};

type commonProps = {
  rows: object[];
  columns: GridColDef[];
  columnsVisibility?: { [key: string]: boolean };
  onSelect?: (
    selectedRows: {
      [key: string]: unknown;
    }[]
  ) => void;
  apiExportEndpoint?: string;
  defaultSelectRowID?: number;
  createNewButton?: JSX.Element;
};

type optionalProps =
  | {
      hidePagination?: false;
      hideFilters?: false;
      /**
       * Map of column field and its corresponding searchMetadata name.
       */
      culumnsMatchingFilterName: Map<string, string>;
      getRows: ActionCreatorWithPreparedPayload<
        [
          {
            [key: string]: unknown;
          }
        ],
        {
          id: string;
          createdAt: string;
          request: GetAction;
        }
      >;
      dispatch: Dispatch;
      queryBuilder: QueryBuilder;
      networkCompletionStack: Record<string, boolean>[];
    }
  | {
      hidePagination: true;
      hideFilters: true;
      culumnsMatchingFilterName?: never;
      getRows?: never;
      dispatch?: never;
      queryBuilder?: never;
      networkCompletionStack?: never;
    }
  | {
      hidePagination: true;
      hideFilters?: false;
      /**
       * Map of column field and its corresponding searchMetadata name.
       */
      culumnsMatchingFilterName: Map<string, string>;
      getRows: ActionCreatorWithPreparedPayload<
        [
          {
            [key: string]: unknown;
          }
        ],
        {
          id: string;
          createdAt: string;
          request: GetAction;
        }
      >;
      dispatch: Dispatch;
      queryBuilder: QueryBuilder;
      networkCompletionStack: Record<string, boolean>[];
    }
  | {
      hidePagination?: false;
      hideFilters: true;
      /**
       * Map of column field and its corresponding searchMetadata name.
       */
      culumnsMatchingFilterName?: never;
      getRows: ActionCreatorWithPreparedPayload<
        [
          {
            [key: string]: unknown;
          }
        ],
        {
          id: string;
          createdAt: string;
          request: GetAction;
        }
      >;
      dispatch: Dispatch;
      queryBuilder: QueryBuilder;
      networkCompletionStack: Record<string, boolean>[];
    };

type propsType = commonProps & optionalProps;

/**
 * NerdGrid component.
 *
 * @param {propsType} props nerdGrid props.
 * @returns {React.ReactElement} DataGrid wrapper with server-side filters.
 */
function NerdGrid(props: propsType): React.ReactElement {
  const {
    columnsVisibility,
    columns,
    networkCompletionStack,
    dispatch,
    queryBuilder,
    getRows,
    culumnsMatchingFilterName,
    hidePagination,
    hideFilters,
    defaultSelectRowID,
    createNewButton,
  } = props;
  const rows: GridRowsProp = props.rows;
  const rowPageOptions = useMemo(() => [10, 50, 100], []);
  let limit: string | undefined;
  let offset: string | undefined;
  let filters: string | undefined;
  if (queryBuilder) {
    const queryParams = queryBuilder.getQuery();
    limit = queryParams["limit"];
    offset = queryParams["offset"];
    filters = queryParams["filters"];
  }
  const [rowsPerPage, setRowsPerPage] = useState(10);
  const [page, setPage] = useState(1);
  const [selectPages, setSelectPages] = useState<number[]>([0]);
  const [disabledNext, setDisabledNext] = useState(false);
  const [rowCount, setRowCount] = React.useState<number | undefined>(undefined);
  const [rowCountState, setRowCountState] = React.useState(rowCount || 0);
  const [queryObject, setQueryObject] = useState<
    | {
        filters: string;
        limit: number;
        offset: number;
      }
    | {
        limit: number;
        offset: number;
      }
  >(
    culumnsMatchingFilterName
      ? {
          filters: filters ? filters : "",
          limit: Number(limit) ? Number(limit) : rowsPerPage,
          offset: Number(offset) ? Number(offset) : 0,
        }
      : {
          limit: Number(limit) ? Number(limit) : rowsPerPage,
          offset: Number(offset) ? Number(offset) : 0,
        }
  );
  const [requestPending, setRequestPending] = useState(false);
  const [requestUUID, setRequestUUID] = useState<string>("");
  const [filterModel, setFilterModel] = useState<GridFilterModel>({
    items: [],
    linkOperator: GridLinkOperator.And,
  });
  const [selectionModel, setSelectionModel] = useState<GridSelectionModel>(
    defaultSelectRowID &&
      rows.filter((row) => row["id"] === defaultSelectRowID).length > 0
      ? [defaultSelectRowID]
      : []
  );
  const prevSelectionModel = React.useRef<GridSelectionModel>(
    defaultSelectRowID ? [defaultSelectRowID] : []
  );

  /**
   * Custom MUI Grid Toolbar
   *
   * @returns {React.ReactElement} Mui Grid Toolbar
   */
  function CustomToolbar(): React.ReactElement {
    return (
      <GridToolbarContainer>
        {!hideFilters && <GridToolbarFilterButton />}
        {props.apiExportEndpoint && (
          <Button
            sx={{ fontSize: "0.8125rem" }}
            variant="text"
            startIcon={<SaveAlt />}
            onClick={() => {
              const a = encodeURI(
                `${props.apiExportEndpoint}${
                  "filters" in queryObject
                    ? `?filters=${queryObject.filters}`
                    : ""
                }`
              );
              window.open(a, "_blank", "noopener,noreferrer");
            }}
          >
            EXPORT
          </Button>
        )}
        {createNewButton}
      </GridToolbarContainer>
    );
  }

  useEffect(() => {
    try {
      if (culumnsMatchingFilterName && filters) {
        const filtersObject = JSON.parse(filters);
        const filterModelTmp: GridFilterModel = {
          items: [],
          linkOperator: GridLinkOperator.And,
        };
        Array.isArray(filtersObject) &&
          filtersObject.forEach(
            (
              filter: { name: string; comparison: string; value: string },
              i: number
            ) => {
              let columnField = "";
              for (const [key, value] of culumnsMatchingFilterName.entries()) {
                if (value === filter.name) columnField = key;
              }
              filterModelTmp.items.push({
                id: i + 1,
                columnField,
                value: filter.value,
                operatorValue: filter.comparison,
              });
            }
          );
        setFilterModel(filterModelTmp);
      }
    } catch (error) {
      return;
    }
  }, []);

  // Set URL query
  useEffect(() => {
    if (queryBuilder) {
      if (
        Number(offset) % Number(limit) !== 0 ||
        !rowPageOptions.includes(Number(limit))
      ) {
        setQueryObject({
          ...queryObject,
          limit: 10,
          offset: 0,
        });
        setPage(1);
        setRowsPerPage(10);
      } else if (Number(limit) === queryObject.limit) {
        setPage(Number(offset) / Number(limit) + 1);
        setRowsPerPage(Number(limit));
      }
      queryBuilder.setUrlQuery(queryObject);
    }
  }, [offset, limit, rowPageOptions, queryObject, queryBuilder]);

  /**
   * Return DataGrid filter operators for a given column type.
   *
   * @param {string | undefined} type Type of the column.
   * @returns {Array<GridFilterOperator>} Return filter operators.
   */
  function operatorsByType(type: string | undefined): GridFilterOperator[] {
    const operators: GridFilterOperator[] = [];
    let input: JSXElementConstructor<GridFilterInputValueProps>;
    switch (type) {
      case "date":
        input = DateOperatorInput;
        break;
      case "boolean":
        input = BoolOperatorInput;
        break;
      default:
        input = StringOperatorInput;
        break;
    }
    Object.entries(OPERATORS).forEach((val) =>
      operators.push({
        label: val[1].label,
        value: val[1].value,
        getApplyFilterFn: () => {
          //ServerSide filter, so we don't need filter function here.
          return (): boolean => {
            return true;
          };
        },
        InputComponent: input,
      })
    );
    return operators;
  }

  const setCustomPagination = useCallback(() => {
    if (Number(offset) > 0 || Number(limit) > 0) {
      setSelectPages((prevState) => [
        ...Array.from(new Set(prevState)),
        ...Array(page)
          .fill(0)
          .map((_e, i) => i + 1),
      ]);
    }
  }, [limit, offset, page]);

  const columnsWithOperators = React.useMemo(
    () =>
      columns.map((col) => {
        return {
          ...col,
          filterOperators: operatorsByType(col.type),
        };
      }),
    [columns]
  );

  useEffect(() => {
    if (dispatch && getRows) {
      setRequestPending(true);
      setRowCount(undefined);
      const uuid = withNetworkCompletionDispatch(
        dispatch,
        getRows(queryObject)
      );
      setRequestUUID(uuid);
    }
  }, [dispatch, getRows, queryObject]);

  // Check if API call ends and close loading
  useEffect(() => {
    if (networkCompletionStack) {
      if (
        requestUUID !== "" &&
        !networkCompletionStack.some((e) => e[requestUUID])
      ) {
        setRequestUUID("");
        setRequestPending(false);
        setRowCount(rows.length);
        //restore selection model after page change
        setSelectionModel(prevSelectionModel.current);
      }
      setCustomPagination();
    }
  }, [requestUUID, networkCompletionStack, setCustomPagination, rows.length]);

  const onFilterChange = React.useCallback(
    (newFilterModel: GridFilterModel) => {
      if (culumnsMatchingFilterName) {
        const filters: { name: string; comparison: string; value: string }[] =
          [];
        newFilterModel.items.forEach((item: GridFilterItem) => {
          const searchMetadataName = culumnsMatchingFilterName.get(
            item.columnField
          );
          if (
            item.operatorValue &&
            item.value !== undefined &&
            item.value !== "" &&
            searchMetadataName
          ) {
            filters.push({
              name: searchMetadataName,
              comparison: item.operatorValue,
              value: item.value,
            });
          }
        });
        setPage(1); // Reset page to 1
        setSelectPages([0]); // Reset select pages
        setQueryObject({
          limit: queryObject.limit,
          offset: 0,
          filters: JSON.stringify(filters),
        });
        setFilterModel(newFilterModel);
      }
    },
    [culumnsMatchingFilterName, queryObject.limit]
  );

  /**
   * Set new page value and change the query object
   *
   * @param {React.MouseEvent<HTMLButtonElement> | null} _event ChangePage
   * @param {number} newPage New page index
   */
  function handleChangePage(
    _event: React.MouseEvent<HTMLButtonElement> | null,
    newPage: number
  ): void {
    prevSelectionModel.current = selectionModel;
    setPage(newPage);
    if (newPage > 0) {
      setQueryObject({
        ...queryObject,
        offset: (newPage - 1) * Number(limit),
      });
    }
  }

  /**
   * @param {ChangeEvent<HTMLInputElement | HTMLTextAreaElement>} event - new rows page number selected
   */
  function handleChangeRowsPerPage(
    event: ChangeEvent<HTMLInputElement | HTMLTextAreaElement>
  ): void {
    const newPage = parseInt(event.target.value, 10);
    setRowsPerPage(newPage); // set new page size
    setPage(1); // Reset page to 1
    setSelectPages([0]); // Reset select pages
    // Set new query object
    setQueryObject((prevState) => ({
      ...prevState,
      limit: newPage,
      offset: 0,
    }));
  }

  /**
   *
   * @param {GridSelectionModel} ids Selected rows
   */
  function onSelectionModelChange(ids: GridSelectionModel) {
    const selectedIDs = new Set(ids);
    const selectedRows = rows.filter((row) => selectedIDs.has(row.id));
    const prevSelectedIDs = new Set(prevSelectionModel.current);
    //Don't call onSelect if page changed
    if (ids.length === 0) {
      if (rows.filter((row) => prevSelectedIDs.has(row.id)).length > 0) {
        props.onSelect?.(selectedRows);
      }
    } else {
      props.onSelect?.(selectedRows);
    }
    setSelectionModel(ids);
  }

  React.useEffect(() => {
    setRowCountState((prevRowCountState) =>
      rowCount !== undefined ? rowCount : prevRowCountState
    );
  }, [rowCount, setRowCountState]);

  return (
    <DataGridPro
      sx={{
        "&.MuiDataGrid-root .MuiDataGrid-cell:focus": {
          outline: "none",
        },
        paddingTop: "0px",
        width: "100%",
        height: "700px",
      }}
      pageSize={rowsPerPage}
      rowsPerPageOptions={rowPageOptions}
      onStateChange={(state: GridState) => {
        const {
          pagination: { rowCount, pageSize },
        } = state;
        setDisabledNext(rowCount < pageSize);
      }}
      pagination
      paginationMode="server"
      rowCount={rowCountState}
      hideFooterPagination={hidePagination}
      components={{
        Toolbar: CustomToolbar,
        Pagination: CustomPagination,
      }}
      componentsProps={{
        filterPanel: { linkOperators: ["and"] },
        pagination: {
          rowsPerPage,
          handleChangeRowsPerPage,
          rowPageOptions,
          handleChangePage,
          page,
          selectPages,
          disabledNext,
        },
      }}
      onSelectionModelChange={onSelectionModelChange}
      rows={rows}
      columns={columnsWithOperators}
      loading={requestPending}
      filterMode="server"
      onFilterModelChange={onFilterChange}
      initialState={{
        columns: {
          columnVisibilityModel: columnsVisibility,
        },
      }}
      filterModel={filterModel}
      selectionModel={selectionModel}
    />
  );
}

export default NerdGrid;
